#!/usr/bin/python
import pdb, re, copy

class Graph:
    OUT=0; IN=1
    EG=0; ND=1
    
    def typeE(elem):
        if type(elem) == tuple and len(elem)==2:
            return Graph.EG
        else:
            return Graph.ND
    typeE = staticmethod(typeE)


    def __init__(self):
        self.elems_pred = [{},{}]
        self.elems = [{}, {}]

    def copy(self):
        tmp = Graph()
        tmp.elems_pred = copy.deepcopy(self.elems_pred)
        tmp.elems = copy.deepcopy(self.elems)
        return tmp

    def translate(self, translation_dict):
        nodes = self.elems[Graph.ND].keys() 
        for node in nodes:
            new_node = translation_dict.get(node, node) 
            self.translateElem(Graph.ND, node, new_node)

        edges = self.elems[Graph.EG].keys() 
        for egde in edges:
            new_edge = (translation_dict.get(edge[0], egde[0]), translation_dict.get(edge[1], egde[1]))
            self.translateElem(Graph.EG, edge, new_edge)
            
    def translateElem(self, tE, old_elem, new_elem):
        if old_elem != new_elem:
            preds = self.elems[tE].pop(old_elem)
            self.elems[tE][new_elem] = preds
            for (pred, val) in preds.items():
                if type(val) in [float, int]:
                    pos = self.elems_pred[tE][pred].index((old_elem,val))
                    self.elems_pred[tE][pred][pos] = (new_elem,val)
                else:
                    pos = self.elems_pred[tE][pred][val].index(old_elem)
                    self.elems_pred[tE][pred][pos][val][pos] = new_elem

    def edgesSet(self):
        return set(self.elems[Graph.EG].keys())

    def nodesSet(self):
        return set(self.elems[Graph.ND].keys())

    def neighborsSub(self, direction = 0, preds=None):
        ndict = {}
        if preds is None:
            preds = self.elems_pred[Graph.EG].keys()

        if direction == -1:
            directions = [0,1]
        else:
            directions = [direction]
        for pred in preds:
            for edge in self.elems_pred[Graph.EG][pred]['EG']:
                for direc in directions:
                    if not ndict.has_key(edge[direc]):
                        ndict[edge[direc]] = set([edge[1-direc]])
                    else:
                        ndict[edge[direc]].add(edge[1-direc])
        return ndict

    def elemPreds(self, elem, preds=None):
        tE = Graph.typeE(elem)
        if preds is None:
            return self.elems[tE].get(elem, {}).items()
        else:
            return [(k,v) for (k,v) in self.elems[tE].get(elem, {}).items() if k in preds]

    def compareNodes(self, nodeA, nodeB, comp_pred=[]):
        comp_set = []
        labels = ['Na=Nb', 'Na<Nb', 'Na>Nb']
        if len(comp_pred) > 0:
            if self.elems[Graph.ND].has_key(nodeA) and self.elems[Graph.ND].has_key(nodeB):
                for pred in comp_pred:
                    if self.elems[Graph.ND][nodeA].has_key(pred) and self.elems[Graph.ND][nodeB].has_key(pred):
                        vS = self.elems[Graph.ND][nodeA][pred]
                        vE = self.elems[Graph.ND][nodeB][pred]
                        comp_set.append((pred, labels[(vS < vE) - (vS > vE)]))
        return comp_set

    def elementMatch(self, matchElem, elem, constraints):
        tE = Graph.typeE(elem)
        match = True
        if constraints.elems[tE].has_key(elem):
            tmp_const = constraints.elems[tE][elem].items()
            while len(tmp_const) > 0 and match:
                pred, val = tmp_const.pop()
                if re.match("^Na[<>=]Nb$", val):
                    match &= (len(self.filterCompare([matchElem], pred, val)) == 1)
                else:
                    ### WARNING: FIRST TEST CAN'T BE MOVED OUT, IN CASE OF COMPARISON EDGES...
                    match &= (self.elems[tE].has_key(matchElem) and self.elems[tE][matchElem].has_key(pred) and self.elems[tE][matchElem][pred]==val)
        return match

    def matchingElements(self, elem, constraints):
        tE = Graph.typeE(elem)
        postproc = []
        tmp_map = None
        for (pred, val) in constraints.elems[tE].get(elem, {}).items():
            if re.match("^Na[<>=]Nb$", val):
                postproc.append((pred, val))
            else:
                tmp_c = set(self.getConstraintMatch(tE, pred, val))
                if tmp_map == None:
                    tmp_map = tmp_c
                else:
                    tmp_map &= tmp_c
        if tmp_map != None:
            tmp_map = list(tmp_map)
            for (pred, val) in postproc:
                tmp_map = self.filterCompare(tmp_map, pred, val)
        return tmp_map

    def filterCompare(self, tmp_map, pred, val):
        labels = ['Na=Nb', 'Na<Nb', 'Na>Nb']
        filtered_list = []
        tE = Graph.ND
        for (nodeA, nodeB) in tmp_map:
            if self.elems[tE][nodeA].has_key(pred) and self.elems[tE][nodeB].has_key(pred):
                vS = self.elems[tE][nodeA][pred]
                vE = self.elems[tE][nodeB][pred]
                if val == labels[(vS < vE) - (vS > vE)] :
                    filtered_list.append((nodeA, nodeB))
        return filtered_list
            
    def getConstraintMatch(self, typeE, pred, val=None): #, filtered=False):
        #### This is not for comparisons!!
        return self.elems_pred[typeE].get(pred, {}).get(val,[])

    def getConstrainedPath(self, path, pred_types):
        node_pred = []; edge_pred = []; comp_pred = []        
        for (k,v) in pred_types[Graph.ND].items():
            if v == "cat":
                node_pred.append(k)
            elif v == "num":
                comp_pred.append(k)
        edge_pred = pred_types[Graph.EG].keys()

        constraintsP = []
        elemi = -1
        for elemi in range(len(path)-1):
            constraintsP.append(self.elemPreds(path[elemi], node_pred))
            constraintsP.append(self.elemPreds((path[elemi], path[elemi+1]), edge_pred))
            constraintsP.append(self.elemPreds((path[elemi+1], path[elemi]), edge_pred))
            for elemj in range(elemi+1):
                constraintsP.append(self.compareNodes(path[elemj], path[elemi+1], comp_pred))
        constraintsP.append(self.elemPreds(path[elemi+1], node_pred))
        return constraintsP

    def readBMG(self, fp, pred_types, patt=None):
        if patt is None:
            patt = ['^(T_)?(?P<NodeOutN>[\w:\d]+)\s(T_)?(?P<NodeInN>[\w:\d]+)\s+(?P<Comm>.*)$', \
                    '^# _attributes (T_)?(?P<NodeN>[\w:\d]+)\s*(?P<Comm>.*)$']
        
        for line in fp:
            tmp = re.match(patt[1], line)
            if tmp is not None:
                for att in tmp.group("Comm").split():
                    parts = att.split("=")
                    if len(parts) == 2:
                        self.parseTriple((tmp.group("NodeN"), parts[0], parts[1]), pred_types)
            else:
                tmp = re.match(patt[0], line)
                if tmp is not None:
                    for att in tmp.group("Comm").split():
                        self.parseTriple((tmp.group("NodeOutN"), att, tmp.group('NodeInN')), pred_types)


    def readTriples(self, fp, pred_types):
        for line in fp:
            self.parseTriple(line.split(), pred_types)

    def parseTriples(self, triples, pred_types):
        for triple in triples:
            self.parseTriple(triple, pred_types)

    def parseTriple(self, triple, pred_types):            
        if len(triple) > 2:
            tE = None
            (x, pred, y) = triple[:3]
            if pred in pred_types[Graph.ND]:
                if pred_types[Graph.ND][pred] == "num":
                    y = float(y)
                tE = Graph.ND
                elem = x
                val = y
                self.parseElem(tE, elem, pred, val)
            elif pred_types[0] is None or pred in pred_types[0]:
                tE = Graph.EG
                elem = (x,y)
                val = 'EG'
                self.parseElem(tE, elem, pred, val)

    def parseElem(self, tE, elem, pred, val):
        if tE == Graph.EG:
            for end_i in [0,1]:
                if not self.elems[Graph.ND].has_key(elem[end_i]):
                    self.elems[Graph.ND][elem[end_i]] = {}
            
        if self.elems[tE].has_key(elem):
            self.elems[tE][elem][pred] = val
        else:
            self.elems[tE][elem] = {pred: val}

        if self.elems_pred[tE].has_key(pred):
            if type(val) in [float, int]:
                self.elems_pred[tE][pred].append((elem, val))                        
            elif self.elems_pred[tE][pred].has_key(val):
                self.elems_pred[tE][pred][val].append(elem)
            else:
                self.elems_pred[tE][pred][val] = [elem]
        else:
            if type(val) in [float, int]:
                self.elems_pred[tE][pred]= [(elem,val)]
            else:
                self.elems_pred[tE][pred]= {val:[elem]}

    def listTriples(self):
        triples = []
        for x, pvs in self.elems[Graph.ND].items():
            for pred, val in pvs.items():
                triples.append((x, pred, val))
                
        for (x,y), preds in self.elems[Graph.EG].items():
            for pred in preds:
                triples.append((x, pred, y))
        return triples

    def writeTriples(self, fp):
        triples = []
        for x, pvs in self.elems[Graph.ND].items():
            for pred, val in pvs.items():
                fp.write("%s %s %s\n" % (x, pred, val))
                
        for (x,y), preds in self.elems[Graph.EG].items():
            for pred, val in preds.items():
                if val != "EG":
                    fp.write("%s %s=%s %s\n" % (x, pred, val, y))
                else:
                    fp.write("%s %s %s\n" % (x, pred, y))

    def writeBMG(self, fp, patt_replace=r'^', replacement='T_', add_att=None):
        for x, pvs in self.elems[Graph.ND].items():
            xf = re.sub(patt_replace, replacement, x)
            fp.write("# _attributes %s " % xf)
            if add_att is not None:
                fp.write(add_att)
            fp.write(" ".join(["%s=%s" % (pred, val) for pred, val in pvs.items()])+"\n")
                
        for (x,y), preds in self.elems[Graph.EG].items():
            xf = re.sub(patt_replace, replacement, x)
            yf = re.sub(patt_replace, replacement, y)
            fp.write("%s %s" % (xf, yf))
            for pred, val in preds.items():
                if val != "EG":
                    fp.write(" %s=%s" % (pred, val))
                else:
                    fp.write(" %s" % pred)
            fp.write('\n')

    def listBMG(self, patt_replace=r'^', replacement='T_', add_att=None):
        lines = []
        for x, pvs in self.elems[Graph.ND].items():
            xf = re.sub(patt_replace, replacement, x)
            lines.append("# _attributes %s " % xf)
            if add_att is not None:
                lines[-1] += add_att
            lines[-1]+= " ".join(["%s=%s" % (pred, val) for pred, val in pvs.items()])
                
        for (x,y), preds in self.elems[Graph.EG].items():
            xf = re.sub(patt_replace, replacement, x)
            yf = re.sub(patt_replace, replacement, y)
            lines.append("%s %s" % (xf, yf))
            for pred, val in preds.items():
                if val != "EG":
                    lines[-1]+= " %s=%s" % (pred, val)
                else:
                    lines[-1]+= " %s" % pred
        return lines

class Data:

    def __init__(self, pred_types = None):
        self.graph = Graph()
        if pred_types is None:
            self.pred_types = [{}, None] # edges, nodes
        else:
            self.pred_types = pred_types
            
    def copy(self):
        tmp = Data()
        tmp.graph = self.graph.copy()
        tmp.pred_types = copy.deepcopy(self.pred_types)
        return tmp

    def writePreds(self, fp):
        for tE in [Graph.EG, Graph.ND]:
            for pred, tP in self.pred_types[tE].items():
                fp.write("##PRED %d %s %s\n" % (tE, pred, tP))

    def parsePreds(self, fp):
        self.pred_types = [{}, {}]
        for line in fp:
            tmp = re.match("^##PRED (?P<tE>[01]) (?P<pred>[\w:\d]+) (?<tP>[a-z]+)$")
            if tmp is not None:
                self.pred_types[tmp.group("tE")][tmp.group("pred")]=tmp.group("tP")

    def readTriples(self, fp):
        self.graph.readTriples(fp, self.pred_types)

    def writeTriples(self, fp):
        self.graph.writeTriples(fp)

    def readBMG(self, fp):
        self.graph.readBMG(fp, self.pred_types)

    def writeBMG(self, fp):
        self.graph.writeBMG(fp)

class NodeMap:

    def __init__(self, mapped_nodes={}):
        self.map_instances = {}
        self.maps_to = {}
        for node, data in mapped_nodes.items():
            self.map_instances[node] = data['pl']
            self.maps_to[node] = set(data['pl'])

    def copy(self):
        tmp = NodeMap()
        tmp.map_instances = copy.deepcopy(self.map_instances)
        tmp.maps_to = copy.deepcopy(self.maps_to)

    def translate(self, translation_dict):
        nodes = self.maps_instances.keys() 
        for node in nodes:
            new_node = translation_dict.get(node, node) 
            if new_node != node:
                self.maps_to[new_node] = self.maps_to.pop(node)
                self.map_instances[new_node] = self.map_instance.pop(node)

    def nbInstances(self):
        return len(self.map_instances.items()[0][1])

    def nbNodes(self):
        return len(self.map_instances)

    def nodesSet(self):
        return set(self.map_instances.keys())
        
    def nodeInstances(self, node):
        return self.map_instances[node]

    def nodeInstance(self, node, pos):
        return self.map_instances[node][pos]

    def allInstances(self):
        return self.map_instances.items()
    
    def nodeMapsTo(self, node):
        return self.maps_to[node]

    def allMapsTo(self):
        return self.maps_to.items()

    def nbMinMapsTo(self):
        return min([ len(mm) for mm in self.maps_to.values()])

class NodeCover:

    BLUE = 0
    RED = 1
    
    def __init__(self, query_nodes=None, nodes_map=None, blues=set(), strict=False):
        self.pairs = [set(), set()]
        self.instances = [[], []]
        self.unordered_instances = set()
        if query_nodes is not None and nodes_map is not None:

            nodes_t = sorted(query_nodes.keys())
            check_relevant = [k for (k,v) in query_nodes.items() if v]

            for j in range(nodes_map.nbInstances()):
                self.unordered_instances.add(tuple(sorted([nodes_map.nodeInstance(node, j) for node in nodes_map.nodesSet()])))
                current_tuple = tuple([nodes_map.nodeInstances(node_t)[j] for node_t in nodes_t])
                if current_tuple in blues:
                    self.instances[NodeCover.BLUE].append(j)
                    self.pairs[NodeCover.BLUE].add(current_tuple)
                else:
                    self.instances[NodeCover.RED].append(j)
                    self.pairs[NodeCover.RED].add(current_tuple)

            if ( strict and blues != self.pairs[NodeCover.BLUE] ): print "Something badly wrong with mining/mapping..."
        
    def copy(self):
        tmp = NodeCover()
        tmp.instances = copy.deepcopy(self.instances)
        tmp.pairs = copy.deepcopy(self.pairs)
        tmp.unordered_instances = copy.deepcopy(self.unordered_instances)
        return tmp

    def bluePairs(self):
        return self.pairs[NodeCover.BLUE]

    def redPairs(self):
        return self.pairs[NodeCover.RED]

    def blueInstances(self):
        return self.instances[NodeCover.BLUE]

    def redInstances(self):
        return self.instances[NodeCover.RED]

    def nbUnordered(self):
        return self.unordered_instances

    def emptyBlue(self):
        self.pairs[NodeCover.BLUE] = set()
        self.instances[NodeCover.BLUE] = []

class Clause:

    def __init__(self):
        self.id = 'TMP'
        self.constraints = None
        self.cover = None
        self.nodes_map = None
        self.query_nodes = None
        self.itemset = None
        self.combination = None

    def copy(self):
        tmp = Clause()
        tmp.id = 'copyOf' + self.getId()
        if self.constraints is not None:
            tmp.constraints = self.constraints.copy()
        if self.query_nodes is not None:
            tmp.query_nodes = copy.deepcopy(self.query_nodes)
        if self.itemset is not None:
            tmp.itemset = copy.deepcopy(self.itemset)
        if self.nodes_map is not None:
            tmp.nodes_map = self.nodes_map.copy()
        if self.cover is not None:
            tmp.cover = self.cover.copy()
        if self.combination is not None:
            tmp.combination = copy.deepcopy(self.combination)

    def translate(self, translation_dict):
        self.constraints.translate(translation_dict)
        self.nodes_map.translate(translation_dict)
        self.translateCombination(translation_dict)

    def translateCombination(self, translation_dict):
        if self.combination is not None and self.combination.has_key('eq_nodes'):
            for i in range(self.combination['eq_nodes']):
                node = self.combination['eq_nodes'][i]
                new_node = translation_dict.get(node, node)
                if node != new_node:
                    self.combination['eq_nodes'][i] = new_node
        
        if self.combination is not None and self.combination.has_key('eq_nodes_counts'):
            nodes =  combined.combination['eq_nodes_counts'].keys()
            for node in nodes:
                new_node = translation_dict.get(node, node) 
                if new_node != node:
                    self.combined.combination['eq_nodes_counts'][new_node] = combined.combination['eq_nodes_counts'].pop(node)



    def getId(self):
        return self.id
    
    def setId(self, nid):
        self.id = nid

    def nbConstraints(self):
        if self.itemset is not None:
            return len(self.itemset)

    def nbBluePairs(self):
        if self.cover is not None:
            return len(self.cover.bluePairs())

    def nbRedPairs(self):
        if self.cover is not None:
            return len(self.cover.redPairs())

    def bluePairs(self):
        if self.cover is not None:
            return self.cover.bluePairs()

    def redPairs(self):
        if self.cover is not None:
            return self.cover.redPairs()

    def blueInstances(self):
        if self.cover is not None:
            return self.cover.blueInstances()

    def redInstances(self):
        if self.cover is not None:
            return self.cover.redInstances()

    def nbUnordered(self):
        if self.cover is not None:
            return self.cover.nbUnordered()

    def blueToRedBlueRatio(self):
        if self.cover is not None:
            return self.cover.nbBluePairs()/float(self.cover.nbBluePairs()+self.cover.nbRedPairs())

    def printBMG(self):
        print '\n'.join(sorted(self.listBMG()))

    def listBMG(self, replace_match='^$', replace_subs='', color_str=""):
        pattern_graph = self.constraints.listBMG(replace_match, replace_subs)

        for node in self.constraints.nodesSet():
            tmp_lab = re.sub(replace_match, replace_subs, node)
            str_atts = color_str
            if self.nodes_map:
                str_atts += "nb_instances=%s nb_mapped=%s " % (self.nodes_map.nbInstances(), len(self.nodes_map.nodeMapsTo(node)) )
            if self.combination is not None:
                str_atts += "nb_eq=%d " % ( self.combination['eq_nodes_counts'].get(node, 1) )
            pattern_graph.append("# _attributes %s %s"  % (tmp_lab, str_atts))
        return pattern_graph
